"""A medley of tools for Hubcap."""

from typing import Iterable, Literal
import time
from functools import partial

from hubcap.base import GithubReader, RepoReader
from hubcap.constants import repo_collection_names
from hubcap.util import RepoSpec, ensure_url_suffix, _prep_git_clone_args


# TODO: Design horribly unclean. Once RepoReader is finished, this should become
# cleaner to write.
def hub(path: RepoSpec):
    path = ensure_url_suffix(path)
    if '/' not in path:
        org = path
        return GithubReader(org)
    # at this point we have at least org/repo/...
    org, repo, *_path = path.split('/')
    if not _path:
        return GithubReader(org)[repo]

    # If not, use RepoReader as the base object  # TODO: Finish RepoReader
    s = RepoReader(f'{org}/{repo}')
    path_iter = iter(_path)
    # TODO: Temporarily commented out -- if not needed, remove
    # if (repo := next(path_iter, None)) is not None:
    #     s = s[repo]
    if (resource := next(path_iter, None)) is not None:
        if resource in repo_collection_names or resource == 'discussions':
            s = RepoReader(s.repo)[resource]
        else:
            # From now we assume the intent is to get a specific branch...
            if resource == 'tree':  # this is to be consistent with browser url access
                # then consider this to be a request for branches
                resource = next(path_iter)
            else:
                # TODO: Change what s is suddenly: Terrible design
                # The point here is that the following instance will then work for
                # getting the branch
                s = GithubReader(org)[repo]
            s = s[resource]

    # Process the rest of the path with the s mapping
    for part in path_iter:
        s = s[part]
    return s


def team_repositories_action(
    repositories: Iterable[str],
    team: str,
    *,
    action: Literal['add_to_repo', 'remove_from_repos'],
    org: str,
    wait_s: int = 1,
):
    """
    Add a list of repositories to a team with read permission
    """
    # Create a GitHub instance using an access token
    g = GithubReader()._github

    # Get the organization object by name
    org_ = g.get_organization(org)

    # Get the team object by name
    team_ = org_.get_team_by_slug(team)
    action_ = getattr(team_, action)

    for repo in repositories:
        # Get the repository object by name
        repo_ = g.get_repo(repo)
        # Carry out the action
        action_(repo_)
        time.sleep(wait_s)


add_repos_to_team = partial(team_repositories_action, action='add_to_repo')
rm_repos_from_team = partial(team_repositories_action, action='remove_from_repos')



r"""
The simplest facade to github data.

Warning: You'll need to have a github api token (google it if you don't have one;
it's easy to get). You'll have to specify this token when making hubcap objects,
or put it in an environmental variable under the name `HUBCAP_GITHUB_TOKEN` or
`GITHUB_TOKEN`


# Example usage

## RepoReader

>>> from hubcap import RepoReader
>>> r = RepoReader('thorwhalen/test_repo')
>>> sorted(r)  # doctest: +SKIP
['artifacts', 'assignees', 'autolinks', 'branches', 'codescan_alerts', 'collaborators', 
'comments', 'commits', 'contributors', 'deployments', 'discussions', 'downloads', 
'environments', 'events', 'forks', 'git_refs', 'hooks', 'issues', 'issues_comments', 
'issues_events', 'labels', 'languages', 'milestones', 'network_events', 
'notifications', 'pending_invitations', 'projects', 'pulls', 'pulls_comments', 
'pulls_review_comments', 'releases', 'repository_advisories', 'secrets', 
'self_hosted_runners', 'stargazers', 'stargazers_with_dates', 'stats_contributors', 
'subscribers', 'tags', 'teams', 'top_paths', 'top_referrers', 'topics', 'variables', 
'watchers', 'workflow_runs', 'workflows']
>>>
>>> 'issues' in r
True
>>> issues = r['issues']
>>> sorted(issues)
[4, 5]
>>> issue_obj = issues[4]
>>> issue_obj
Issue(title="Test Issue A", number=4)
>>> sorted([attr for attr in dir(issue_obj) if not attr.startswith('_')])  # doctest: +SKIP
['CHECK_AFTER_INIT_FLAG', 'active_lock_reason', 'add_to_assignees', 'add_to_labels', 
'as_pull_request', 'assignee', 'assignees', 'body', 'closed_at', 'closed_by', 
'comments', 'comments_url', 'create_comment', 'create_reaction', 'created_at', 
'delete_labels', 'delete_reaction', 'edit', 'etag', 'events_url', 'get__repr__', 
'get_comment', 'get_comments', 'get_events', 'get_labels', 'get_reactions', 
'get_timeline', 'html_url', 'id', 'labels', 'labels_url', 'last_modified', 'lock', 
'locked', 'milestone', 'number', 'pull_request', 'raw_data', 'raw_headers', 
'remove_from_assignees', 'remove_from_labels', 'repository', 'setCheckAfterInitFlag', 
'set_labels', 'state', 'state_reason', 'title', 'unlock', 'update', 'updated_at', 
'url', 'user']
>>> issue_obj.number
4
>>> issue_obj.state
'open'
>>> issue_obj.labels
[Label(name="documentation"), Label(name="enhancement")]
>>> # title of issue
>>> issue_obj.title
'Test Issue A'
>>> # content of issue
>>> issue_obj.body
'Contents of Test Issue A'
>>>
>>> issue_obj.comments
2
>>>
>>> list(issue_obj.get_comments())  # doctest: +NORMALIZE_WHITESPACE
[IssueComment(user=NamedUser(login="thorwhalen"), id=1801792378), 
IssueComment(user=NamedUser(login="thorwhalen"), id=1801792855)]
>>> issue_comment = issue_obj.get_comment(1801792378)
>>> issue_comment.body
'Comment 1 of Test Issue A'
>>>
>>> 'discussions' in r
True
>>> discussions = r['discussions']  # doctest: +SKIP
>>> sorted(discussions)  # doctest: +SKIP
[1, 2, 3]

## hub function

The high level function `hub` is the simplest way to get started. It's a
function that takes a path to a github resource and returns a mapping to that
resource. The mapping is lazy, so it's only when you access a key that the
resource is actually fetched from github.

>>> repositories = hub('thorwhalen')
>>> 'hubcap' in repositories
True
>>> hubcap_repo = repositories['hubcap']
>>>
>>> repo = hub('thorwhalen/hubcap')
>>> 'master' in repo
True
>>> master_files = repo['master']
>>>
>>> files = hub('thorwhalen/hubcap/master')
>>> '/README.md' in files
True
>>> '/hubcap/' in files
True
>>> hubcap_files = files['/hubcap/']
>>> '/hubcap/base.py' in hubcap_files
True

### Access issues

>>> issues = hub('thorwhalen/hubcap/issues')
>>> 3 in issues  # there's a "number 3" issue
True
>>> issue = issues[3]
>>> issue.title
'Test Issue'
>>> issue.body
'This is just a test issue to test that hubcap can see it.\r\n'
>>> issue.comments  # meaning "number of comments"
2

### Access discussions

>>> discussions = hub('thorwhalen/hubcap/discussions')  # doctest: +SKIP
>>> 2 in discussions  # doctest: +SKIP
True
>>> discussion = discussions[2]    # doctest: +SKIP
>>> discussion['title']    # doctest: +SKIP
'Root interface of hubcap'


## GithubReader

One of the main classes is `GithubReader`. It's a mapping that connects to a
github user or organization, and returns a mapping of repositories. The
repositories are also mappings, that return mappings of branches, and so on.

>>> from hubcap import GithubReader
>>> s = GithubReader('thorwhalen')  # connnecting to a particular user/organization
>>> list(s)  # doctest: +SKIP
['agen',
 'aix',
 ...
 'viral',
 'wealth',
 'wrapt']
>>> 'a_non_existing_repository_name' in s
False
>>> 'hubcap' in s  # of course, this will be true, it's what you're using now!
True
>>> repo = s['hubcap']
>>> list(repo)
['master']
>>> branch = repo['master']
>>> list(branch)  # doctest: +NORMALIZE_WHITESPACE
['/.gitattributes', '/.github/', '/.gitignore', '/LICENSE',
'/README.md', '/docsrc/', '/hubcap/', '/misc/', '/setup.cfg', '/setup.py']
>>> content = branch['/setup.cfg']
>>> print(content[:32].decode())
[metadata]
name = hubcap
version

>>> from hubcap import get_repository_info
>>> info = get_repository_info('thorwhalen/hubcap')
>>> list(info)  # doctest: +NORMALIZE_WHITESPACE
['name', 'full_name', 'description', 'stargazers_count',
'forks_count', 'watchers_count', 'html_url', 'last_commit_date']
>>> info['name']
'hubcap'
>>> info['html_url']
'https://github.com/thorwhalen/hubcap'
>>> info['stargazers_count'] >= 1
True

You also have other useful objects, like `Issues`, `IssueComments`, `Discussions`, etc.   


"""

from github import GithubException, Github, ContentFile
from hubcap.base import (
    RepoReader,
    GithubReader,
    find_github_token,
    Branches,
    BranchDir,
    Issues,
    IssueComments,
    GitHubReader,  # backcompatibility alias of GithubReader
)
from hubcap.util import (
    get_repository_info,
    cached_github_object,
    Discussions,
    git_clone,
    git_wiki_clone,
    create_markdown_from_jdict,  # Creates a markdown representation of a discussion (metadata json-dict).
    replace_relative_urls,  # replace relative urls with absolute ones
    parse_github_url,  #  parse a GitHub URL and returns a dict of its components
    generate_github_url,  # generate a GitHub URL from the provided components dict.
    transform_github_url,  # transform a GitHub URL to another type, updating components as needed.
)
from hubcap.tools import hub
from hubcap.repo_slurp import repo_text_aggregate



"""Tools to slurp a repo's information """

# TODO: This module is a bit of a mess. It should be cleaned up and reorganized.
#  It should also be tested.
#  There's some light caching functionality for files and wiki
#  (via explicitly giving a folder as the repo) but it is not consistent.
#  Better use graze, or something like it, for cachine.
#  What we'd want is to be able to just speak the language of urls, and let the module
#  take care of caching what it clones (or in the case of discussions, downloads)

import os
from typing import Mapping, Union, Literal
from hubcap.util import (
    ensure_github_url,
    git_clone,
    git_wiki_clone,
)
from hubcap.base import RepoReader, ResourceNotFound, NotSet


# -------------------------------------------------------------------------------------
# markdown utils

from typing import Callable, KT, VT, Iterable

KvToText = Callable[[KT, VT], str]


def kv_to_python_aware_markdown(k, v):
    """Puts .py content in a code block, and returns a markdown formatted string."""
    if k.endswith('.py'):
        # if it's a python file, we'll put it in a code block
        v = f'```python\n{v}\n```\n'
    return f'## {k}\\n\\n{v}\\n\\n'


def _text_segments_from_mapping(
    store, kv_to_text: KvToText = kv_to_python_aware_markdown
) -> Iterable[str]:
    """
    Generates text segments from a given store (mapping) using a formatting function.

    Args:
        store (dict): A mapping with string keys and values.
        kv_to_text (callable): A function that takes a key and value and returns a formatted string.

    Yields:
        str: Text for each kv item
    """
    for key, value in store.items():
        if value is not None:  # None values are skipped (can use None as sentinel)
            yield kv_to_text(key, value)


def text_from_mapping(
    mapping: Mapping, kv_to_text: KvToText = kv_to_python_aware_markdown
):
    """
    Creates a string from a given mapping with string keys and values using a formatting function.

    Args:
        mapping (dict): A mapping with string keys and values.
        kv_to_text (callable): Function that takes a key and value and returns a string.

    Returns:
        str: A string aggregate of the values of the mapping.

    Example:
    >>> example_mapping = {
    ...     "apple.txt": "Crumble",
    ...     "tit.md": "For tat.",
    ...     "state.cfg": "Of the art."
    ... }
    >>> example_kv_to_text = lambda k, v: f"## {k}\\n```\\n{v}\\n```"
    >>> print(text_from_mapping(example_mapping, example_kv_to_text))
    ## apple.txt
    ```
    Crumble
    ```
    ## tit.md
    ```
    For tat.
    ```
    ## state.cfg
    ```
    Of the art.
    ```
    """
    return '\n'.join(_text_segments_from_mapping(mapping, kv_to_text))


def add_offset_to_headers(markdown_text: str, offset: int = 0) -> str:
    r"""
    Returns the same text but where the (markdown) headers have been offset by `offset`.

    It works by adding `offset` number of '#' to the start of each header line.

    If the offset causes a header level to become zero or negative, it is ignored (remains the same).

    Comments (lines starting with '#') in code blocks are not affected by the offset.

    >>> add_offset_to_headers("## Header 1\n\n### Header 2\n\n#### Header 3", 1)
    '### Header 1\n\n#### Header 2\n\n##### Header 3'
    >>> add_offset_to_headers("## Header 1\n\n### Header 2\n\n#### Header 3", -1)
    '# Header 1\n\n## Header 2\n\n### Header 3'

    Note that comments in code blocks are not affected:

    >>> add_offset_to_headers("```\n# Not a header\n```", 1)
    '```\n# Not a header\n```'

    >>> add_offset_to_headers("## Header 1\n\n```\n# Not a header\n```\n\n### Header 2", 1)
    '### Header 1\n\n```\n# Not a header\n```\n\n#### Header 2'
    """

    def offset_header_line(line: str):
        if line.startswith('#'):
            header_level = line.count('#', 0, line.find(' '))
            new_level = header_level + offset
            if new_level > 0:
                return '#' * new_level + line[header_level:]
            else:
                return line
        return line

    in_code_block = False

    def _result_lines():
        nonlocal in_code_block
        for line in markdown_text.split('\n'):
            if line.startswith('```'):
                in_code_block = not in_code_block
                yield line
            elif in_code_block:
                yield line
            else:
                yield offset_header_line(line)

    return '\n'.join(_result_lines())


def _ensure_callable_processor(processor, if_true=lambda x: x, if_false=lambda x: None):
    """
    Ensures that the given processor is a callable. If the processor is a boolean,
    it returns a default callable based on the boolean value.

    Args:
        processor (bool or callable): The processor to ensure is callable.
        if_true (callable): The callable to use if the processor is True.
        if_false (callable): The callable to use if the processor is False.

    Returns:
        callable: The ensured callable processor.

    Raises:
        ValueError: If the processor is neither a boolean nor a callable.

    Examples:
        >>> _ensure_callable_processor(True)("test")
        'test'

        >>> _ensure_callable_processor(False)("test")

        >>> _ensure_callable_processor(lambda x: x.upper())("test")
        'TEST'

        >>> _ensure_callable_processor("not callable")
        Traceback (most recent call last):
            ...
        ValueError: not callable is not a callable
    """
    if processor is True:
        return if_true
    elif processor is False:
        return if_false
    elif callable(processor):
        return processor
    else:
        raise ValueError(f'{processor} is not a callable')


def _markdown_lines(notebook, process_code, process_markdown, process_output):
    """Helper function to yield markdown lines from a notebook."""
    process_code = _ensure_callable_processor(process_code)
    process_markdown = _ensure_callable_processor(process_markdown)
    process_output = _ensure_callable_processor(process_output)

    for cell in notebook.cells:
        if cell.cell_type == 'markdown':
            processed = process_markdown(cell.source)
            if processed is not None:
                yield processed
        elif cell.cell_type == 'code':
            processed_code = process_code(cell.source)
            if processed_code is not None:
                yield f'```python\n{processed_code}\n```'
            for output in cell.get('outputs', []):
                if output.output_type == 'stream':
                    text = output.text
                elif output.output_type == 'error':
                    text = ''.join(output.traceback)
                elif output.output_type == 'execute_result':
                    text = output.data.get('text/plain', '')
                else:
                    continue
                processed_output = process_output(text)
                if processed_output is not None:
                    yield f'```\n{processed_output}\n```'


# TODO: Move to markdown utils module or package
# TODO: Write a few useful process_* functions to get useful markdown from code and output cells
#   For example, not including traceback in error outputs, or only including the last line of
#   output cells -- or not including scrap sections of code cells.
def notebook_to_markdown(
    notebook: Union[str, bytes],
    *,
    process_code=True,
    process_markdown=True,
    process_output=True,
    encoding='utf-8',
):
    """
    Transforms a Jupyter notebook into a markdown string with control over cell processing.

    Args:
        notebook: Path to (or bytes or str contents of) the Jupyter notebook.
        process_code (bool or callable): Whether to include code cells or a callable to process them.
                                         Default is True.
        process_markdown (bool or callable): Whether to include markdown cells or a callable to process them.
                                             Default is True.
        process_output (bool or callable): Whether to include the output of code cells or a callable to process them.
                                           Default is True.

    Returns:
        str: The notebook content as a markdown string.
    """
    import nbformat
    import io

    # Make a notebook object
    if os.path.isfile(notebook):
        with open(notebook, 'r', encoding=encoding) as f:
            notebook = nbformat.read(f, as_version=4)
    if isinstance(notebook, (bytes, str)):
        if isinstance(notebook, str):
            notebook = io.StringIO(notebook)
        elif isinstance(notebook, bytes):
            notebook = io.BytesIO(notebook)
        else:
            raise ValueError(f'Unsupported type for notebook: {type(notebook)}')
        notebook = nbformat.read(notebook, as_version=4)
    else:
        assert isinstance(notebook, nbformat.NotebookNode)

    return '\n\n'.join(
        filter(
            None,
            _markdown_lines(notebook, process_code, process_markdown, process_output),
        )
    )


# -------------------------------------------------------------------------------------
# github_repo_text_aggregate

import tempfile
import re
from functools import partial
from subprocess import CalledProcessError
from dol import TextFiles, wrap_kvs, filt_iter, Files


def _is_local_git_repo(repo: str):
    return os.path.isdir(repo) and os.path.isdir(os.path.join(repo, '.git'))


def ensure_repo_folder(repo: str, clone_func=git_clone):
    """Returns a local repo folder.
    Either the input repo is already a local "git" folder (which we return as is),
    or a temporary folder to clone the repo into is returned.
    """
    if os.path.isdir(repo):
        return repo
    elif ensure_github_url(repo):
        try:
            local_folder = tempfile.mkdtemp()
            clone_func(repo, local_folder)
            return local_folder
        except CalledProcessError:
            raise ResourceNotFound(
                f"Couldn't find a (or error cloning with {clone_func}) a resource at {repo}"
            )
    else:
        raise ResourceNotFound(f"Couldn't find a local git repo at {repo}")


def _decode_to_text_or_skip(obj, log_error_function=False):
    try:
        return obj.decode('utf-8')
    except UnicodeDecodeError:
        if log_error_function:
            # if log_error_function
            print(f'Error decoding {obj}')
        return None  # Note: None values will be skipped in text_from_mapping


def all_decodable_text_folder_to_mapping(folder):
    """A folder_to_mapping function that takes"""
    return wrap_kvs(Files(folder), obj_of_data=_decode_to_text_or_skip)


def key_filtered_text_files(folder, key_pattern):
    return filt_iter(TextFiles(folder), filt=re.compile(key_pattern).search)


_pattern_for_python_and_markdown_files = r'.*\.(py|md)$'

_filtered_py_and_md_files = partial(
    key_filtered_text_files, key_pattern=_pattern_for_python_and_markdown_files
)


def _does_not_start_with_docsrc_or_setup(key: KT):
    return not key.startswith('docsrc/') and not key.startswith('setup.')


# TODO: A lot more can be done to parametrize the construction of a discussion text
#   if and when more format flexibility is needed (for example, to enable parsing
#  (text-back-to-json) or a nicer text rendering of the discussion)
#  For formatting, may want to detect headers in content and modify according to
#  header_level.
def __discussion_json_to_text_segments(
    d: dict, header_level=2, offset_headers=True
) -> Iterable[str]:
    if offset_headers:
        if offset_headers is True:
            offset = header_level
        elif isinstance(offset_headers, int):
            offset = offset_headers
        else:
            raise ValueError(
                f'offset_headers must be an int or True, not {offset_headers}'
            )
        _offset_headers = partial(add_offset_to_headers, offset=offset)
    else:
        _offset_headers = lambda x: x

    yield _offset_headers(d['body']) + '\n\n'
    for comment_num, comment in enumerate(d['comments'], 1):
        yield '#' * (header_level + 1) + f' Comment {comment_num}\n\n'
        yield _offset_headers(comment['body']) + '\n\n'
        for reply in comment.get('replies', []):
            yield '#' * (header_level + 2) + f' Reply\n\n'
            yield _offset_headers(reply['body'] + '\n\n')


def _discussion_json_to_text(d: dict, header_level=2):
    return ''.join(__discussion_json_to_text_segments(d, header_level))


def discussions_mapping(repo, discussion_json_to_text=_discussion_json_to_text):
    repo_url = ensure_github_url(repo)
    discussions = RepoReader(repo_url)['discussions']

    def _gen():
        for key, d in discussions.items():
            yield f"Discussion {key}: {d['title']}", discussion_json_to_text(d)

    return dict(_gen())


def wiki_mapping(repo, local_repo_folder=None, default=NotSet):
    try:
        local_repo_folder = ensure_repo_folder(repo, clone_func=git_wiki_clone)
        s = filt_iter(TextFiles(local_repo_folder), filt=re.compile(r'.*\.md$').search)
        return s
    except ResourceNotFound:
        if default is NotSet:
            raise
        return default


def repo_files_mapping(
    repo: str,
    *,
    kv_to_text: KvToText = kv_to_python_aware_markdown,
    folder_to_mapping: Union[Callable[[str], Mapping], str] = _filtered_py_and_md_files,
    extra_key_filter=_does_not_start_with_docsrc_or_setup,
):
    local_repo_folder = ensure_repo_folder(repo)

    # If the folder_to_mapping is a string, use the key_filtered_text_files function
    # with the string as the key pattern
    if isinstance(folder_to_mapping, str):
        key_filter_pattern = folder_to_mapping
        folder_to_mapping = partial(
            key_filtered_text_files, key_pattern=key_filter_pattern
        )

    # Create a mapping for the files in the target folder
    mapping = folder_to_mapping(local_repo_folder)

    if extra_key_filter:
        if isinstance(extra_key_filter, str):
            extra_key_filter = re.compile(extra_key_filter).search
        mapping = filt_iter(mapping, filt=extra_key_filter)

    mapping = wrap_kvs(mapping, postget=kv_to_text)

    return mapping


CloneKinds = Literal['files', 'wiki', 'discussions']


def github_repo_mapping(
    repo: str, *, kind: CloneKinds = 'files', repo_files_mapping=repo_files_mapping,
):
    r"""
    Clone a git repository and make a mapping of the files in the repository.

    Args:
        repo (str): The GitHub repository in the format 'owner/repo' or github url.
        clone_to_folder (str, optional): The target folder to clone the repository into. Defaults to a temporary folder.

    Returns:
        Mapping: A mapping of the files in the repository.


    >>> repo_files = github_repo_mapping('thorwhalen/hubcap')  # doctest: +SKIP

    """
    # repo = ensure_github_url(repo)

    # subprocess.check_call(clone_command, shell=True)
    if kind == 'discussions':
        return discussions_mapping(repo)
    elif kind == 'wiki':
        return wiki_mapping(repo, default={})
    elif kind == 'files':
        return repo_files_mapping(repo)
    else:
        raise ValueError(f"kind must be 'files' or 'wiki', not {kind}")


def repo_text_aggregate(
    repo,
    kinds: Union[CloneKinds, Iterable[CloneKinds]] = ('files', 'wiki', 'discussions'),
    *,
    github_repo_mapping=github_repo_mapping,
    text_from_mapping=text_from_mapping,
):
    """
    Clone a git repository and aggregate all file contents into a string.

    Args:
        repo (str): The GitHub repository in the format 'owner/repo' or github url.
        kinds (Union[CloneKinds, Iterable[CloneKinds]], optional):
            The kinds of content to aggregate. Defaults to ('files', 'wiki', 'discussions').

    Returns:
        str: A string with all file contents.


    >>> aggregate = repo_text_aggregate('thorwhalen/hubcap')  # doctest: +SKIP
    >>> print(aggregate[:100])  # doctest: +SKIP
    ## README.md\n\n# hubcap
    A [dol](https://github.com/i2mint/dol) (i.e. dict-like) interface to github...

    """
    text = ''
    for kind in kinds:
        mapping = github_repo_mapping(repo, kind=kind)
        text += text_from_mapping(mapping)
    return text



"""Download wikis from github"""

import os
import subprocess
import shutil
import tempfile


def download_github_wiki(repo_id: str, download_dir: str):
    """
    Download the wiki associated with a given GitHub repository.

    Parameters
    ----------
    repo_id : str
        The GitHub repository identifier in the format "owner/repository_name".
    download_dir : str
        The path of the directory where the wiki will be downloaded.

    Raises
    ------
    FileNotFoundError
        If the download_dir does not exist.
    """

    if not os.path.isdir(download_dir):
        os.makedirs(download_dir)

    owner, repo_name = repo_id.split('/')
    destination = os.path.join(download_dir, owner, repo_name)

    # GitHub stores wiki content in a separate git repository appended with ".wiki"
    wiki_repo = f'https://github.com/{repo_id}.wiki.git'

    with tempfile.TemporaryDirectory() as tmpdir:
        try:
            subprocess.check_call(['git', 'clone', wiki_repo, tmpdir])
        except subprocess.CalledProcessError:
            # If git clone fails, wiki probably does not exist, so return
            return

        # If we get here, wiki exists and we've cloned it successfully to a temporary directory.
        # Now we move the wiki files to the destination directory.

        os.makedirs(destination, exist_ok=True)
        for file_name in os.listdir(tmpdir):
            shutil.move(os.path.join(tmpdir, file_name), destination)


# def download_github_wiki(repo_id: str, download_dir: str):
#     """
#     Download all wikis of a GitHub repository to a specific directory.

#     Parameters
#     ----------
#     repo_id : str
#         Identifier of the repository in "owner/repository_name" format.
#     download_dir : str
#         Directory where wikis will be downloaded.
#     """
#     owner, repo_name = repo_id.split("/")
#     dest_dir = os.path.join(download_dir, owner, repo_name)

#     # Create the destination directory if it doesn't exist
#     os.makedirs(dest_dir, exist_ok=True)

#     # URL of the wiki's Git repository
#     wiki_git_url = f"https://github.com/{repo_id}.wiki.git"

#     try:
#         # Clone the wiki Git repository
#         subprocess.run(
#             ["git", "clone", wiki_git_url, dest_dir],
#             stderr=subprocess.DEVNULL,
#             check=True,
#         )
#     except subprocess.CalledProcessError:
#         # If the repository doesn't have a wiki, remove the created directory
#         os.rmdir(dest_dir)
#     except Exception as e:
#         print(f"An error occurred when downloading the wiki of {repo_id}: {e}")


def url_of_git_directory(path: str) -> str:
    """
    Check if a directory is a git repository and return its remote URL.

    Parameters
    ----------
    path : str
        The path of the directory to check.

    Returns
    -------
    str
        The remote URL if the directory is a git repository, None otherwise.
    """
    try:
        # Get the URL of the 'origin' remote
        with open(os.devnull, 'w') as devnull:
            url = subprocess.check_output(
                ['git', '-C', path, 'remote', 'get-url', 'origin'],
                stderr=devnull,
                encoding='utf-8',
            ).strip()
        return url
    except subprocess.CalledProcessError:
        return None
    except Exception as e:
        print(f'An error occurred when checking if {path} is a git directory: {e}')
        return None


def repository_identifiers(rootdir: str, recursive: bool = True):
    """
    Yield GitHub repository URLs for all git repositories under a given directory.

    Parameters
    ----------
    rootdir : str
        The root directory from where to start the search.
    recursive : bool, optional
        If True, search recursively in all subdirectories.
        If False, search only in the root directory.
        By default, True.

    Yields
    ------
    str
        The GitHub repository URL for each git repository.
    """
    for dirpath, dirnames, filenames in os.walk(rootdir):
        url = url_of_git_directory(dirpath)
        if url is not None:
            # This is a git repository, yield the repository URL
            # and stop exploring this directory further
            if url:
                yield url_to_repository_identifier(url)
            dirnames.clear()  # This will prevent os.walk from exploring subdirectories
        elif not recursive:
            # If we are not searching recursively, clear the dirnames list
            # so os.walk does not explore subdirectories
            dirnames.clear()


def url_to_repository_identifier(url: str):
    if isinstance(url, str):
        if url.endswith('.git'):
            url = url[:-4]  # Remove .git from the end
        if url.startswith('https://github.com/'):
            return url[len('https://github.com/') :]



"""A place to put constants, defaults, types..."""


from typing import Any, Literal, Tuple, Dict, Iterable, Union, Callable, NewType
from datetime import datetime

from github.Repository import Repository
from dol.signatures import Sig

# --------------------------------------------------------------------------- #
# Some functions to help create the constants


def _non_callable_non_dundered_attrs(obj: Any) -> Tuple[str]:
    return tuple(
        x for x in dir(obj) if not x.startswith('_') and not callable(getattr(obj, x))
    )


def _non_callable_non_dundered_attrs_of_repo_type() -> Tuple[str]:
    return _non_callable_non_dundered_attrs(Repository)


def _last_commit_date(repo: Repository):
    return (
        datetime.strptime(
            repo.updated_at.strftime('%Y-%m-%d %H:%M:%S'), '%Y-%m-%d %H:%M:%S'
        )
        .date()
        .isoformat()
    )


# --------------------------------------------------------------------------- #

# RepoProperty is a string that is a valid Repository property
repo_props = _non_callable_non_dundered_attrs_of_repo_type()
RepoProperty = Literal[repo_props]
RepoFunc = Callable[[Repository], Any]  # A function whose input is a Repository
RepoPropSpec = Union[RepoProperty, RepoFunc]  # A repo property or function on repo
# The normal way is to specify RepoInfo with a dict,
#  but iterable of items allows us to define a immutable default repo_info, so:
RepoInfo = Union[
    Dict[str, RepoProperty],  # {field: prop_spec, ...}
    Iterable[Tuple[str, RepoProperty]],  # [(field, prop_spec), ...]
    str,  # "prop1 prop2 ..."
]

_dflt_repo_props = 'name full_name description stargazers_count forks_count watchers_count html_url'.split()

DFLT_REPO_INFO = tuple(
    [*zip(_dflt_repo_props, _dflt_repo_props), ('last_commit_date', _last_commit_date)]
)


def repo_collection_names_():
    """
    List of the names of the objects that can be retrieved from a repository
    (e.g. commits, contributors, issues, pull_requests, releases, tags, etc.)
    These names are extracted by taking all the method names of the
    `github.Repository` class that start with `get_` and end with `s`,
    and removing the `get_` prefix.
    """
    repo_object_names = {
        k for k in dir(Repository) if k.startswith('get_') and k.endswith('s')
    }
    repo_object_names = {
        k[len('get_') :]
        for k in repo_object_names
        if Sig(getattr(Repository, k)).n_required == 1
    }
    return repo_object_names - set(dir(dict))


# tuple.__doc__ is read-only, so had to subclass to give my variable a doc
_tuple = type('_tuple', (tuple,), {'__doc__': repo_collection_names_.__doc__})
repo_collection_names = _tuple(sorted(repo_collection_names_()))

# TODO: When in 3.11, change to Literal[*repo_collection_names]
RepoCollectionNames = Literal[repo_collection_names]  # type: ignore



"""Base objects"""

from functools import cached_property, partial
from operator import attrgetter
from typing import Mapping
from dol import KvReader, wrap_kvs
from dol.util import format_invocation

import github
from github import GithubException
from github.Auth import Token, Login

from hubcap.util import (
    RepoSpec,
    ensure_repo_obj,
    Github,
    Repository,
    Discussions,
)
from hubcap.constants import repo_collection_names


NotSet = github.GithubObject.NotSet


class ResourceNotFound(Exception):
    """Raised when a resource is not found"""


class RepositoryNotFound(github.UnknownObjectException, ResourceNotFound):
    """Raised when a repository is not found"""


def decoded_contents(content_file):
    return content_file.decoded_content
    # from base64 import b64decode
    # return b64decode(content_file.content).decode()


# TODO: Enable user parametrization of how to find token
#  Use: config2py
def find_github_token():
    import os

    return os.environ.get('HUBCAP_GITHUB_TOKEN', None) or os.environ.get(
        'GITHUB_TOKEN', None
    )


def find_user_name():
    import os

    return os.environ.get('HUBCAP_GITHUB_TOKEN', None) or os.environ.get(
        'GITHUB_USERNAME', None
    )


# --------------------------------------------------------------------------------------
# RepoObjects


def identity(x):
    return x


class RepoObjects(KvReader):
    def __init__(
        self,
        repo: RepoSpec,
        get_objs,
        *,
        objs_to_items=enumerate,
        data_of_obj=identity,
        get_objs_kwargs=None,
    ):
        self.repo = ensure_repo_obj(repo)
        self.get_objs = get_objs
        self.data_of_obj = data_of_obj
        self.get_objs_kwargs = dict(get_objs_kwargs or {})

        if isinstance(objs_to_items, str):
            key_attr = objs_to_items
            _get_key = attrgetter(key_attr)
            self.objs_to_items = lambda x: zip(map(_get_key, x), x)
        else:
            assert callable(
                objs_to_items
            ), 'issue_objs_to_itemskey must be a str or callable'
            self.objs_to_items = objs_to_items

    @cached_property
    def _objs(self):
        return {
            k: v
            for k, v in self.objs_to_items(
                self.get_objs(self.repo, **self.get_objs_kwargs)
            )
        }

    def __iter__(self):
        yield from self._objs

    def __getitem__(self, k):
        return self.data_of_obj(self._objs[k])

    # def __repr__(self):
    #     return f"{type(self).__name__}({self.repo}, {self.get_objs}, ...)"


from hubcap.util import repo_collections_configs

dflt_repo_kwargs = {
    k: {'objs_to_items': v} for k, v in repo_collections_configs.items()
}
dflt_repo_kwargs['issues']['get_objs_kwargs'] = (('state', 'open'),)


def repo_objects_instance(repo, object_name: str) -> Mapping:
    if object_name == 'issues':
        return Issues(repo)
    elif object_name == 'discussions':
        return Discussions(repo)
    elif object_name in repo_collection_names:
        return RepoObjects(
            repo,
            get_objs=getattr(Repository, f'get_{object_name}'),
            **dflt_repo_kwargs.get(object_name, {}),
        )
    else:
        raise KeyError(f'Unknown object name: {object_name}')


class RepoReader(KvReader):
    repo_collection_names = sorted(set(repo_collection_names) | {'discussions'})

    # TODO: Separate error handling concern: https://github.com/i2mint/i2/issues/45
    def __init__(self, repo: RepoSpec):
        try:
            self.repo = ensure_repo_obj(repo)
        except github.UnknownObjectException as e:
            if next(iter(e.args), None) == 404:
                raise RepositoryNotFound(f'Repository not found: {repo}')
            else:
                raise

    def __getitem__(self, k):
        return repo_objects_instance(self.repo, k)

    def __iter__(self):
        yield from self.repo_collection_names

    def __contains__(self, k):
        return k in self.repo_collection_names

    def __repr__(self):
        return f'{type(self).__name__}("{self.repo.full_name}")'


# Extras -----------------------------------------------------------------------------


class IssueCommentsBase(KvReader):
    """Base Mapping interface to repository issue comments (object)"""

    _comment_key = enumerate

    def __init__(self, issue_obj):
        self.src = issue_obj

    @cached_property
    def _comments(self):
        return {k: v for k, v in self._comment_key(self.src.get_comments())}

    def __iter__(self):
        yield from self._comments

    def __getitem__(self, k):
        return self._comments[k]


@wrap_kvs(obj_of_data=attrgetter('body'))
class IssueComments(IssueCommentsBase):
    """Mapping interface to repository issue comments' body"""


class IssueContents(KvReader):
    """Mapping interface to repository issue contents"""

    def __init__(self, issue_obj):
        self.src = issue_obj

    def __iter__(self):
        yield from ('body', 'comments')

    def __getitem__(self, k):
        if k == 'comments':
            return IssueComments(self.src)
        else:
            return self.src.body

    def __repr__(self):
        return format_invocation(self.__class__.__name__, (self.src,))


# TODO: Figure out how to make Issues and Workflows (pickable) classes automatically
#   (The instances can be made from repo_objects_instance)
class Issues(RepoObjects):
    """
    Mapping interface to repository issues.

    :param repo: The repository to get issues from
    :param get_objs: The function to get the issues
    :param objs_to_items: A function to get the key-value pairs from the issue objects
    :param data_of_obj: A function to get the data from the issue object.
        Default is attrgetter('body'). Use identity to get the whole issue object.
    :param get_objs_kwargs: The keyword arguments to pass to the get_objs function

    The default of data_of_obj is `IssueContents`, which will give a mapping
    interface to the body and comments of the issue.
    """

    def __init__(
        self,
        repo: RepoSpec,
        get_objs=Repository.get_issues,
        *,
        objs_to_items='number',
        data_of_obj=identity,  # IssueContents,
        get_objs_kwargs=(('state', 'open'),),
    ):
        super().__init__(
            repo,
            get_objs=get_objs,
            objs_to_items=objs_to_items,
            data_of_obj=data_of_obj,
            get_objs_kwargs=get_objs_kwargs,
        )


class Workflows(RepoObjects):
    def __init__(
        self,
        repo: RepoSpec,
        get_objs=Repository.get_workflows,
        *,
        objs_to_items='id',
        data_of_obj=identity,
        get_objs_kwargs=(),
    ):
        super().__init__(
            repo,
            get_objs=get_objs,
            objs_to_items=objs_to_items,
            data_of_obj=data_of_obj,
            get_objs_kwargs=get_objs_kwargs,
        )


# --------------------------------------------------------------------------------------
# GithubReader


# TODO: use signature arithmetic
# @kv_decorator
class GithubReader(KvReader):
    """
    a Store that can access a GitHub account.

    You need to specify a token in auth, or in the environment variable 
    HUBCAP_GITHUB_TOKEN or GITHUB_TOKEN.

    The iteration is defined to be the repositories of the account_name 
    (could be organization name). 
    """

    def __init__(
        self,
        account_name: str = None,
        content_file_extractor=decoded_contents,
        *,
        auth=None,
        jwt=None,
        base_url='https://api.github.com',
        timeout=15,
        user_agent='PyGithub/Python',
        per_page=30,
        verify=True,
        retry=None,
        get_repos_kwargs=(),
        login_or_token=None,  # deprecated
        password=None,  # deprecated
    ):
        # TODO: Not sure what the rules are with account_name and login_or_token
        #  So only using a search-if-not-given strategy if account_name is None
        # account_name = account_name or find_user_name()

        if auth is None and login_or_token is not None:
            if password is not None:
                auth = Login(login_or_token, password)
            else:
                token = login_or_token
                auth = Token(token)
        else:
            auth = auth or find_github_token()

        if auth is not None:
            # TODO: This forces token use. Should we allow for login/password use?
            auth = Token(auth)
        else:
            assert isinstance(
                account_name, str
            ), 'account_name must be given (and a str)'

        _github = Github(
            auth=auth,
            login_or_token=login_or_token,
            password=password,
            jwt=jwt,
            base_url=base_url,
            timeout=timeout,
            user_agent=user_agent,
            per_page=per_page,
            verify=verify,
            retry=retry,
        )
        self._github = _github
        self.src = (
            _github.get_user(account_name) if account_name else _github.get_user()
        )
        self.content_file_extractor = content_file_extractor
        self.get_repos_kwargs = dict(get_repos_kwargs)

    def __iter__(self):
        for x in self.src.get_repos(**self.get_repos_kwargs):
            org, name = x.full_name.split('/')
            if org == self.src.login:
                yield name

    def __getitem__(self, k):
        """Retrieves a given repository
        :param k: str
        :rtype: :class:`github.Repository.Repository`
        """
        try:
            repository = self.src.get_repo(k)
        except GithubException as e:
            raise KeyError(f"Key doesn't exist: {k}")
        return Branches(repository, self.content_file_extractor)

    def __contains__(self, k):
        if self.get(k, None) is None:
            return False
        return True

    def __repr__(self):
        return format_invocation(
            self.__class__.__name__, (self.src, self.content_file_extractor),
        )


GitHubReader = GithubReader  # backcompatibility alias


class Branches(KvReader):
    def __init__(self, repo: RepoSpec, content_file_extractor=decoded_contents):
        self.src = ensure_repo_obj(repo)
        self.content_file_extractor = content_file_extractor
        # self._con = repo  # same as this.

    def __iter__(self):
        yield from (x.name for x in self.src.get_branches())

    def __getitem__(self, k):
        # return self.src.get_branch(k) # should not give only the branch
        # return self.src.get_contents("", ref = k)
        return BranchDir(
            self.src,
            branch_name=k,
            path='',
            content_file_extractor=self.content_file_extractor,
        )

    def __repr__(self):
        return format_invocation(
            self.__class__.__name__, (self.src, self.content_file_extractor),
        )


class BranchDir(KvReader):
    def __init__(
        self,
        repo: RepoSpec,
        branch_name,
        path='',
        content_file_extractor=decoded_contents,
    ):
        self.src = ensure_repo_obj(repo)
        self.branch_name = branch_name
        self.path = path
        self.content_file_extractor = content_file_extractor

    def __iter__(self):
        yield from (
            self.path + '/' + x.name + ('/' if x.type == 'dir' else '')
            for x in self.src.get_contents(self.path, ref=self.branch_name)
        )
        # yield from (x.name for x in self.src.get_contents(self.subpath, ref=self.branch_name))

    def __getitem__(self, k):
        if k.endswith('/'):
            k = k[:-1]  # remove the / suffix
        t = self.src.get_contents(k)
        # TODO: There is an inefficiency here in the isinstance(t, list) case
        if isinstance(
            t, list
        ):  # TODO: ... you already have the content_files in t, so don't need to call API again.
            return self.__class__(
                self.src, self.branch_name, k, self.content_file_extractor,
            )
        else:
            return self.content_file_extractor(t)

    def __repr__(self):
        return format_invocation(
            self.__class__.__name__,
            (self.src, self.branch_name, self.path, self.content_file_extractor,),
        )
        # return f"{self.__class__.__name__}({self.src}, {self.branch_name})"


# --------------------------------------------------------------------------------------
# Not used, but for principle:


def _content_file_isfile(content_file):
    return content_file.type == 'file'


def _content_file_isdir(content_file):
    return content_file.type == 'dir'


# from dol import kv_wrap
#
# BranchContent = kv_wrap.outcoming_vals(lambda x: x.contents if isinstance(x, ))

from github import PaginatedList


class PaginatedListDol:
    """Gives a Sized & Iterable & Reversible view of a paginated_list.
    That is, iterating over a PaginatedListDol
    """

    def __init__(self, paginated_list: PaginatedList):
        self.paginated_list = paginated_list

    def __iter__(self):
        for item in self.paginated_list:
            yield item

    def __len__(self):
        return self.paginated_list.totalCount

    def __reversed__(self):
        return PaginatedListDol(self.paginated_list.reversed)


class GitHubDol(KvReader):
    """
    a Store that can access a GitHub account
    """

    def __init__(
        self,
        account_name: str = None,
        content_file_extractor=decoded_contents,
        login_or_token=None,
        password=None,
        jwt=None,
        base_url='https://api.github.com',
        timeout=15,
        user_agent='PyGithub/Python',
        per_page=30,
        verify=True,
        retry=None,
        pool_size=None,
    ):
        assert isinstance(account_name, str), 'account_name must be given (and a str)'
        login_or_token = login_or_token or find_github_token()

        _github = Github(
            login_or_token=login_or_token,
            password=password,
            jwt=jwt,
            base_url=base_url,
            timeout=timeout,
            user_agent=user_agent,
            per_page=per_page,
            verify=verify,
            retry=retry,
            pool_size=pool_size,
        )
        self._github = _github
        self.src = (
            _github.get_user(account_name) if account_name else _github.get_user()
        )
        self.content_file_extractor = content_file_extractor

    def __iter__(self):
        for x in self.src.get_repos():
            org, name = x.full_name.split('/')
            if org == self.src.login:
                yield name

    def __getitem__(self, k):
        """Retrieves a given repository
        :param k: str
        :rtype: :class:`github.Repository.Repository`
        """
        try:
            repository = self.src.get_repo(k)
        except GithubException as e:
            raise KeyError(f"Key doesn't exist: {k}")
        return Branches(repository, self.content_file_extractor)

    def __contains__(self, k):
        if self.get(k, None) is None:
            return False
        return True

    def __repr__(self):
        return format_invocation(
            self.__class__.__name__, (self.src, self.content_file_extractor),
        )
